# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  2014 Sep 29, Cornelius Kölbel, cornelius@privacyidea.org
#               Added userlist test and PEP8 stuff
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
Description:  funcitonal tests
                
Dependencies: -
'''

import logging
from privacyidea.tests import TestController, url

log = logging.getLogger(__name__)


class TestAdminController(TestController):

    def set_tmp_policy(self, action, user="*"):
        parameters = {'name': 'tmp_admin_pol',
                      'scope': 'admin',
                      'realm': '*',
                      'action': action,
                      'user': user,
                      'selftest_admin': 'superadmin'
                      }

        response = self.app.get(url(controller='system',
                                    action='setPolicy'),
                                params=parameters)
        self.assertTrue('"setPolicy tmp_admin_pol"' in response, response)
        
    def del_tmp_policy(self, action):
        parameters = {'name': 'tmp_admin_pol',
                      'selftest_admin': 'superadmin'
                      }

        response = self.app.get(url(controller='system',
                                    action='delPolicy'),
                                params=parameters)
        self.assertTrue('"status": true' in response, response)
        self.assertTrue('delPolicy' in response, response)
        
    def createToken3(self):
        parameters = {"serial": "003e808e",
                      "otpkey": "e56eb2bcbafb2eea9bce9463f550f86d587d6c71",
                      "description": "my EToken",
                      }

        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)

    def createToken2(self, serial="F722362"):
        parameters = {"serial": serial,
                      "otpkey": "AD8EABE235FC57C815B26CEF3709075580B44738",
                      "description": "TestToken" + serial,
                      }

        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)
        return serial

    def createTokenSHA256(self, serial="SHA256"):
        parameters = {"serial": serial,
                      "otpkey": "47F6EE05C06FA1CDB8B9AADF520FCF86221DB0A"
                                "107731452AE140EED0EB518B0",
                      "type": "hmac",
                      "hashlib": "sha256"
                      }
        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)
        return serial

    def createToken(self):
        parameters = {"serial": "F722362",
                      "otpkey": "AD8EABE235FC57C815B26CEF3709075580B44738",
                      "user": "root",
                      "pin": "pin",
                      "description": "TestToken1",
                      }

        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)

        parameters = {"serial": "F722363",
                      "otpkey": "AD8EABE235FC57C815B26CEF"
                                "3709075580B4473880B44738",
                      "user": "root",
                      "pin": "pin",
                      "description": "TestToken2",
                      }

        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)

        parameters = {"serial": "F722364",
                      "otpkey": "AD8EABE235FC57C815B26CEF37090755",
                      "user": "root",
                      "pin": "pin",
                      "description": "TestToken3",
                      }

        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)

        # test the update
        parameters = {"serial": "F722364",
                      "otpkey": "AD8EABE235FC57C815B26CEF37090755",
                      "user": "root",
                      "pin": "Pin3",
                      "description": "TestToken3",
                      }

        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)

    def removeTokenBySerial(self, serial):

        parameters = {"serial": serial,
                      }

        response = self.app.get(url(controller='admin', action='remove'),
                                params=parameters)
        return response

    def removeTokenByUser(self, user):
        # final delete all tokens of user root
        parameters = {"user": user,
                      }

        response = self.app.get(url(controller='admin', action='remove'),
                                params=parameters)
        return response

    def showToken(self):
        response = self.app.get(url(controller='admin', action='show'))
        return response

    def test_0000_000(self):
        self.deleteAllTokens()

    def test_set(self):
        self.createToken()

        parameters = {"serial": "F722364",
                      "pin": "pin",
                      "MaxFailCount": "20",
                      "SyncWindow": "400",
                      "OtpLen": "6",
                      "hashlib": "sha256"
                      }

        response = self.app.get(url(controller='admin', action='set'),
                                params=parameters)
        self.assertTrue('"set pin": 1' in response, response)
        self.assertTrue('"set SyncWindow": 1' in response, response)
        self.assertTrue('"set OtpLen": 1' in response, response)
        self.assertTrue('"set MaxFailCount": 1' in response, response)
        self.assertTrue('"set hashlib": 1' in response, response)

        parameters = {"user": "root",
                      "pin": "pin",
                      "MaxFailCount": "20",
                      "SyncWindow": "400",
                      "OtpLen": "6",
                      }

        response = self.app.get(url(controller='admin', action='set'),
                                params=parameters)
        self.assertTrue('"set pin": 3' in response, response)
        self.assertTrue('"set SyncWindow": 3' in response, response)
        self.assertTrue('"set OtpLen": 3' in response, response)
        self.assertTrue('"set MaxFailCount": 3' in response, response)

        response = self.removeTokenBySerial("F722362")
        self.assertTrue('"value": 1' in response, response)
        response = self.removeTokenByUser("root")
        self.assertTrue('"value": 2' in response, response)

    def test_remove(self):
        self.createToken()
        response = self.removeTokenByUser("root")
        log.debug(response)

    def test_enable(self):
        self.createToken()
        parameters = {"serial": "F722364"}
        response = self.app.get(url(controller='admin', action='disable'),
                                params=parameters)
        self.assertTrue('"value": 1' in response, response)

        parameters = {"serial": "F722364"}
        response = self.app.get(url(controller='admin', action='show'),
                                params=parameters)

        self.assertTrue('false' in response, response)
        self.assertTrue('F722364' in response, response)

        parameters = {"serial": "F722364"}
        response = self.app.get(url(controller='admin', action='enable'),
                                params=parameters)
        self.assertTrue('"value": 1' in response, response)

        parameters = {"serial": "F722364"}
        response = self.app.get(url(controller='admin', action='show'),
                                params=parameters)

        self.assertTrue('true' in response, response)
        self.assertTrue('F722364' in response, response)

        self.removeTokenByUser("root")
        
    def test_show_csv(self):
        '''
        testing token csv export
        '''
        response = self.app.get(url(controller='admin', action='show'),
                                params={"outform": "csv"})
        print response.headers
        assert(response.headers.get("Content-Disposition") ==
               "attachment; filename=privacyidea-tokendata.csv")
        
    def test_show_policy_fail(self):
        '''
        testing show with failing policy
        '''
        self.set_tmp_policy("show", user="superadmin")
        response = self.app.get(url(controller='admin', action='show'),
                                params={'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have any rights in any realm! "
               "Check the policies." in response)
        self.del_tmp_policy("show")

    def test_show_with_realm(self):
        '''
        testing admin/show with filterrealm
        '''
        response = self.app.get(url(controller='system', action='getRealms'),
                                params=None)
        self.assertTrue("mydefrealm" in response, response)
        
        response = self.app.get(url(controller='admin', action='show'),
                                params={"viewrealm": "mydefrealm"})
        self.assertTrue('"status": true,' in response, response)
        
    def test_token_realm(self):
        '''
        testing setting the tokenrealm
        '''
        response = self.app.get(url(controller="admin", action="init"),
                                params={"serial": "TR_001",
                                        "type": "spass"})
        self.assertTrue('"status": true,' in response, response)
        
        # successful set token realm
        response = self.app.get(url(controller="admin", action="tokenrealm"),
                                params={"serial": "TR_001",
                                        "realms": "mydefrealm, myotherrealm"})
        self.assertTrue('"status": true,' in response, response)
        
        # set tokenrealm of token, that does not exist.
        response = self.app.get(url(controller="admin", action="tokenrealm"),
                                params={"serial": "TR_001XXX",
                                        "realms": "mydefrealm, myotherrealm"})
        self.assertTrue('No token with serial TR_001XXX found' in response,
                        response)
        
    def test_reset(self):
        response = self.app.get(url(controller="admin", action="init"),
                                params={"serial": "RE_001",
                                        "type": "hmac",
                                        "otpkey": "123456123456123456"})
        self.assertTrue('"status": true,' in response, response)
        
        response = self.app.get(url(controller="admin", action="reset"),
                                params={"serial": "RE_001"})
        self.assertTrue('"status": true,' in response, response)
        self.assertTrue('"value": 1' in response, response)
        
        response = self.app.get(url(controller="admin", action="reset"),
                                params={"serial": "RE_001XXXX"})
        self.assertTrue('"status": true,' in response, response)
        self.assertTrue('"value": 0' in response, response)
        
    def test_losttoken(self):
        response = self.app.get(url(controller="admin", action="init"),
                                params={"serial": "LOST_001",
                                        "type": "hmac",
                                        "otpkey": "123456123456123456"})
        self.assertTrue('"status": true,' in response, response)
        
        response = self.app.get(url(controller="admin", action="losttoken"),
                                params={"serial": "LOST_001"})
        self.assertTrue('"status": false,' in response, response)
        self.assertTrue('You can only define a lost token for'
                        ' an assigned token.' in response, response)
        
        response = self.app.get(url(controller="admin", action="assign"),
                                params={"serial": "LOST_001",
                                        "user": "root"})
        self.assertTrue('"status": true,' in response, response)
        
        response = self.app.get(url(controller="admin", action="losttoken"),
                                params={"serial": "LOST_001"})
        self.assertTrue('"serial": "lostLOST_001",' in response, response)
        self.assertTrue('"password": "' in response, response)
        
        response = self.app.get(url(controller="admin", action="unassign"),
                                params={"serial": "LOST_001"})
        self.assertTrue('"status": true,' in response, response)
        
        response = self.app.get(url(controller="admin", action="remove"),
                                params={"serial": "LOST_001"})
        self.assertTrue('"status": true,' in response, response)
        
    def test_userlist(self):
        response = self.app.get(url(controller='admin', action='userlist'),
                                params={"username": "*"})
        self.assertTrue("passthru_user2" in response, response)
        self.assertTrue("passthru_user1" in response, response)
        self.assertTrue("remoteuser" in response, response)
        self.assertTrue("user1" in response, response)
        self.assertTrue("user2" in response, response)
        
        response = self.app.get(url(controller='admin', action='userlist'),
                                params={"username": "p*"})
        self.assertTrue('"username": "passthru_user1"' in response, response)
        self.assertTrue('"username": "passthru_user2"' in response, response)
        self.assertTrue('"username": "remoteuser"' not in response, response)
        self.assertTrue('"username": "user1"' not in response, response)
        self.assertTrue('"username": "user2"' not in response, response)
        
    def test_unassign_policy_fail(self):
        self.set_tmp_policy("unassign", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='unassign'),
                                params={'serial': 'S123456',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right to "
               "unassign token S123456." in response)
        self.del_tmp_policy("unassign")

    def test_resync(self):

        self.createToken()

        # test resync of token 2
        parameters = {"user": "root", "otp1": "359864", "otp2": "348449"}
        response = self.app.get(url(controller='admin', action='resync'),
                                params=parameters)

        self.assertTrue('"value": false' in response, response)

        parameters = {"user": "root",
                      "otp1": "359864",
                      "otp2": "348448"}
        response = self.app.get(url(controller='admin', action='resync'),
                                params=parameters)
        # Test response...
        log.error("response %s\n", response)
        self.assertTrue('"value": true' in response, response)

        self.removeTokenBySerial("F722364")
        self.removeTokenBySerial("F722363")
        self.removeTokenBySerial("F722362")

    def test_resync_sha256(self):
        self.createTokenSHA256(serial="SHA256")

        parameters = {"serial": "SHA256",
                      "otp1": "778729",
                      "otp2": "094573"}
        response = self.app.get(url(controller="admin", action="resync"),
                                params=parameters)

        self.assertTrue('"value": true' in response, response)
        self.removeTokenBySerial("SHA256")

    def test_setPin(self):
        self.createToken3()

        # test resync of token 2
        parameters = {"serial": "003e808e",
                      "userpin": "123456",
                      "sopin": "123234"}
        response = self.app.get(url(controller='admin', action='setPin'),
                                params=parameters)
        # log.error("response %s\n",response)
        # Test response...
        self.assertTrue('"set sopin": 1' in response, response)
        self.assertTrue('"set userpin": 1' in response, response)

        self.removeTokenBySerial("003e808e")
        
    def test_setpin_policy_fail(self):
        self.set_tmp_policy("setPin", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='setPin'),
                                params={'serial': 'S123456',
                                        'userpin': '1111',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right to set "
               "MOTP PIN/ SC UserPIN for token S123456" in response)
        self.del_tmp_policy("setPin")
        
    def test_assign(self):

        serial = self.createToken2(serial="F722362")

        # response = self.app.get(url(controller='admin', action='show'))

        respRealms = self.app.get(url(controller='system', action='getRealms'),
                                  params=None)
        log.debug(respRealms)

        # test initial assign
        parameters = {"serial": serial,
                      "user": "root"}
        response = self.app.get(url(controller='admin', action='assign'),
                                params=parameters)
        # log.error("response %s\n",response)
        # Test response...
        self.assertTrue('"value": true' in response, response)

        # test initial assign update
        parameters = {"serial": serial, "user": "root", "pin": "NewPin"}
        response = self.app.get(url(controller='admin', action='assign'),
                                params=parameters)
        # Test response...
        self.assertTrue('ERR1103: Token already assigned to user' in response,
                        response)
        
        parameters = {"serial": serial}
        response = self.app.get(url(controller='admin', action='unassign'),
                                params=parameters)
        
        parameters = {"serial": serial, "user": "root", "pin": "NewPin"}
        response = self.app.get(url(controller='admin', action='assign'),
                                params=parameters)
        # Test response...
        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='show'))

        self.assertTrue('"User.userid": "0", ' in response, response)

        # test initial assign update
        parameters = {"serial": serial, "user": "root"}
        response = self.app.get(url(controller='admin', action='unassign'),
                                params=parameters)
        # log.error("response %s\n",response)
        self.assertTrue('"value": true' in response, response)

        # test wrong assign
        parameters = {"serial": serial, "user": "NoBody"}
        response = self.app.get(url(controller='admin', action='assign'),
                                params=parameters)
        # log.error("response %s\n",response)
        self.assertTrue('getUserId failed: no user '
                        '>NoBody< found!' in response, response)

        response = self.app.get(url(controller='admin', action='show'))

        self.assertTrue('"User.userid": "",' in response, response)

        self.removeTokenBySerial(serial)

    def test_assign_policy_fail(self):
        self.set_tmp_policy("assign", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='assign'),
                                params={'serial': 'S123456',
                                        'user': 'root',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right" in response)
        self.del_tmp_policy("assign")
        
    def test_init_policy_fail(self):
        self.set_tmp_policy("initHMAC", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='init'),
                                params={'serial': 'S123456',
                                        'otpkey': '1234',
                                        'type': 'hmac',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right" in response)
        self.del_tmp_policy("initHMAC")
        
    def test_set_policy_fail(self):
        self.set_tmp_policy("set", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='set'),
                                params={'serial': 'S123456',
                                        'phone': '1234',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right" in response)
        self.del_tmp_policy("set")
        
    def test_resync_policy_fail(self):
        self.set_tmp_policy("resync", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='resync'),
                                params={'serial': 'S123456',
                                        'otp1': '1234',
                                        'otp2': '1234',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right" in response)
        self.del_tmp_policy("resync")
        
    def test_userlist_policy_fail(self):
        self.set_tmp_policy("userlist", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='userlist'),
                                params={'username': '*',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right" in response)
        self.del_tmp_policy("userlist")
        
    def test_tokenrealm_policy_fail(self):
        self.set_tmp_policy("tokenrealm", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='tokenrealm'),
                                params={'serial': 'S123456',
                                        'realms': 'r1,r2,r3',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right" in response)
        self.del_tmp_policy("tokenrealm")
        
    def test_reset_policy_fail(self):
        self.set_tmp_policy("reset", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='reset'),
                                params={'serial': 'S123456',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right" in response)
        self.del_tmp_policy("reset")
        
    def test_copytokenpin_policy_fail(self):
        self.set_tmp_policy("copytokenpin", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='copyTokenPin'),
                                params={'from': 'S123456',
                                        'to': 'S234567',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right" in response)
        self.del_tmp_policy("copytokenpin")

    def test_copytokenuser_policy_fail(self):
        self.set_tmp_policy("copytokenuser", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='copyTokenUser'),
                                params={'from': 'S123456',
                                        'to': 'S234567',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right" in response)
        self.del_tmp_policy("copytokenuser")

    def test_losttoken_policy_fail(self):
        self.set_tmp_policy("losttoken", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='losttoken'),
                                params={'serial': 'S123456',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right" in response)
        self.del_tmp_policy("losttoken")
        
    def test_checkstatus_policy_fail(self):
        self.set_tmp_policy("checkstatus", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='checkstatus'),
                                params={'serial': 'S123456',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right" in response)
        self.del_tmp_policy("checkstatus")
        
    def test_checkstatus(self):
        response = self.app.get(url(controller='admin',
                                    action='checkstatus'),
                                params={'serial': 'S123456',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"values": {}' in response
        
        response = self.app.get(url(controller='admin',
                                    action='checkstatus'),
                                params={'selftest_admin': 'looser_admin'})
        print response
        assert 'ERR77' in response
        assert 'admin/checkstatus: check the token status' in response
        
        response = self.app.get(url(controller='admin',
                                    action='checkstatus'),
                                params={'user': 'root',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"values": {}' in response

    def test_assign_umlaut(self):
        self.createTokenSHA256(serial="umlauttoken")

        parameters = {"serial": "umlauttoken", "user": "kölbel"}
        response = self.app.get(url(controller="admin", action="assign"),
                                params=parameters)

        self.assertTrue('"value": true' in response, response)

        self.removeTokenBySerial("umlauttoken")

    def test_enroll_umlaut(self):

        parameters = {"serial": "umlauttoken",
                      "otpkey": "47F6EE05C06FA1CDB8B9AADF520FCF86"
                                "221DB0A107731452AE140EED0EB518B0",
                      "type": "hmac",
                      "hashlib": "sha256",
                      "user": "kölbel"
                      }
        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters)
        self.assertTrue('"value": true' in response, response)
        self.removeTokenBySerial("umlauttoken")

    def test_login(self):
        '''
        Testing login and logout
        account/dologin?login=corny%40r2&realm=r1&password=test
        '''
        
        response = self.app.get("http://localhost:5001/account/dologin",
                                params={'login': 'admin2@admin',
                                        'realm': '',
                                        'password': 'secret'})
        
        self.assertTrue('302 Found' in response.body, response.body)
        self.assertTrue('you should be redirected automatically.'
                        in response.body, response.body)
        
        response = self.app.get("http://localhost:5001/account/logout",
                                params={})

        print response.body
        self.assertTrue('302 Found' in response.body, response.body)
        self.assertTrue('you should be redirected automatically.'
                        in response.body, response.body)

    def test_check_serial(self):
        '''
        Checking what happens if serial exists
        '''
        response = self.app.get(url(controller='admin', action='init'),
                                params={"serial": 'unique_serial_001',
                                        "type": 'spass'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='check_serial'),
                                params={'serial': 'unique_serial_002'})

        self.assertTrue('"unique": true' in response, response)
        self.assertTrue('"new_serial": "unique_serial_002"' in response,
                        response)

        response = self.app.get(url(controller='admin', action='check_serial'),
                                params={'serial': 'unique_serial_001'})

        self.assertTrue('"unique": false' in response, response)
        self.assertTrue('"new_serial": "unique_serial_001_01"' in response,
                        response)
        
    def test_checkserial_policy_fail(self):
        self.set_tmp_policy("checkserial", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='check_serial'),
                                params={'serial': 'S123456',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right to"
               " run checkserial. Check the policies." in response)
        self.del_tmp_policy("checkserial")

    def test_setPin_empty(self):
        '''
        Testing setting empty PIN and SO PIN
        '''
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'setpin_01',
                                        'type': 'spass'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='setPin'),
                                params={'serial': 'setpin_01'})

        self.assertTrue('"status": false' in response, response)
        self.assertTrue('"code": 77' in response, response)

        response = self.app.get(url(controller='admin', action='setPin'),
                                params={'serial': 'setpin_01',
                                        'sopin': 'geheim'})

        self.assertTrue('"set sopin": 1' in response, response)

    def test_set_misc(self):
        '''
        Setting CountWindow, timeWindow, timeStep, timeShift
        '''
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'token_set_misc',
                                        'type': 'spass'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='set'),
                                params={'serial': 'token_set_misc',
                                        'CounterWindow': '100',
                                        'timeWindow': '180',
                                        'timeStep': '30',
                                        'timeShift': '0'})

        self.assertTrue('set CounterWindow": 1' in response, response)
        self.assertTrue('"set timeShift": 1' in response, response)
        self.assertTrue('"set timeWindow": 1' in response, response)
        self.assertTrue('"set timeStep": 1' in response, response)

    def test_set_count(self):
        '''
        Setting countAuth, countAuthMax, countAuthSucces countAuthSuccessMax
        '''
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'token_set_count',
                                        'type': 'spass'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='set'),
                                params={'serial': 'token_set_count',
                                        'countAuth': '10',
                                        'countAuthMax': '180',
                                        'countAuthSuccess': '0',
                                        'countAuthSuccessMax': '10'})

        self.assertTrue('"set countAuthSuccess": 1' in response, response)
        self.assertTrue('"set countAuthSuccessMax": 1' in response, response)
        self.assertTrue('"set countAuth": 1' in response, response)
        self.assertTrue('"set countAuthMax": 1' in response, response)

        return

    def test_set_validity(self):
        '''
        Setting validity period
        '''
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'token_set_validity',
                                        'type': 'spass'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='set'),
                                params={'serial': 'token_set_validity',
                                        'validityPeriodStart': '2012-10-12',
                                        'validityPeriodEnd': '2013-12-30',
                                        })

        self.assertTrue('"status": false' in response, response)
        self.assertTrue('does not match format' in response, response)

        params = {'serial': 'token_set_validity',
                  'validityPeriodStart': '12/12/12 10:00',
                  'validityPeriodEnd': '30/12/13 13:00',
                  }
        response = self.app.get(url(controller='admin', action='set'),
                                params=params)

        self.assertTrue('"status": true' in response, response)
        self.assertTrue('"set validityPeriodStart": 1' in response, response)
        self.assertTrue('"set validityPeriodEnd": 1' in response, response)

    def test_set_empty(self):
        '''
        Running set without parameter
        '''
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'token_set_empty',
                                        'type': 'spass'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='set'),
                                params={'serial': 'token_set_empty',
                                        })

        self.assertTrue('"status": false' in response, response)
        self.assertTrue('"code": 77' in response, response)

    def test_copy_token_pin(self):
        '''
        testing copyTokenPin

        We create one token with a PIN and authenticate.
        Then we copy the PIN to another token and try to authenticate.
        '''
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'copy_token_1',
                                        'type': 'spass',
                                        'pin': '1234'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='validate', action='check_s'),
                                params={'serial': 'copy_token_1',
                                        'pass': '1234'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'copy_token_2',
                                        'type': 'spass',
                                        'pin': 'otherPassword'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='validate', action='check_s'),
                                params={'serial': 'copy_token_2',
                                        'pass': 'otherPassword'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='copyTokenPin'),
                                params={'from': 'copy_token_1',
                                        'to': 'copy_token_2'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='validate', action='check_s'),
                                params={'serial': 'copy_token_2',
                                        'pass': '1234'})

        self.assertTrue('"value": true' in response, response)

    def test_copy_token_user(self):
        '''
        testing copyTokenUser
        '''
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'copy_user_1',
                                        'type': 'spass',
                                        'pin': 'copyTokenUser',
                                        'user': 'root'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='validate', action='check'),
                                params={'user': 'root',
                                        'pass': 'copyTokenUser'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'copy_user_2',
                                        'type': 'spass',
                                        'pin': 'unknownSecret'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin',
                                    action='copyTokenUser'),
                                params={'from': 'copy_user_1',
                                        'to': 'copy_user_2'})

        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='validate', action='check'),
                                params={'user': 'root',
                                        'pass': 'unknownSecret'})

        self.assertTrue('"value": true' in response, response)

    def test_enroll_token_twice(self):
        '''
        test to enroll another token with the same serial number
        '''
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'token01',
                                        'type': 'hmac',
                                        'otpkey': '123456'})

        self.assertTrue('"value": true' in response, response)

        # enrolling the token of the same type is possible
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'token01',
                                        'type': 'hmac',
                                        'otpkey': '567890'})

        self.assertTrue('"value": true' in response, response)

        # enrolling of another type is not possible
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'token01',
                                        'type': 'spass',
                                        'otpkey': '123456'})

        self.assertTrue("already exist with type" in response, response)
        self.assertTrue("Can not initialize token with new type" in response,
                        response)

        # clean up
        response = self.app.get(url(controller='admin', action='remove'),
                                params={'serial': 'token01'})

        self.assertTrue('"status": true' in response, response)

    def test_serial_by_otp(self):
        '''
        Test gettoken/multiotp and admin/getSerialByOtp
        '''
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'gsbo001',
                                        'type': 'hmac',
                                        'otpkey': '12345678901234'
                                                  '567890123456789012',
                                        'user': 'root',
                                        'selftest_admin': 'superadmin'})
        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='system', action='setPolicy'),
                                params={'name': 'gettoken',
                                        'scope': 'gettoken',
                                        'realm': 'myDefRealm',
                                        'action': 'max_count_hotp=20'})
        self.assertTrue('"setPolicy gettoken"' in response, response)
        
        response = self.app.get(url(controller='gettoken',
                                    action='getmultiotp'),
                                params={'serial': 'gsbo001',
                                        'count': '10'})
        self.assertTrue('"3": "180096"' in response, response)
        
        response = self.app.get(url(controller='admin',
                                    action='getSerialByOtp'),
                                params={'otp': '180096'})
        self.assertTrue('"serial": "gsbo001"' in response, response)
        
        response = self.app.get(url(controller='admin', action='remove'),
                                params={'serial': 'gsbo001'})
        self.assertTrue('"value": 1' in response, response)
        
    def test_getserialbyotp_policy_fail(self):
        self.set_tmp_policy("getserial", user="superadmin")
        response = self.app.get(url(controller='admin',
                                    action='getSerialByOtp'),
                                params={'otp': '180096',
                                        'selftest_admin': 'looser_admin'})
        print response
        assert '"status": false' in response
        assert("ERR410: You do not have the administrative right to get"
               " serials by OTPs in this realm!" in response)
        self.del_tmp_policy("getserial")
        
    def test_setting_description(self):
        '''
        Test setting the description and phone of a token
        '''
        response = self.app.get(url(controller='admin', action='init'),
                                params={'serial': 'desc001',
                                        'type': 'hmac',
                                        'otpkey': '1234567890'
                                                  '1234567890123456789012',
                                        'selftest_admin': 'superadmin'})
        self.assertTrue('"value": true' in response, response)
        
        response = self.app.get(url(controller='admin', action='set'),
                                params={'serial': 'desc001',
                                        'description': 'something',
                                        'phone': '1234560'})
        
        self.assertTrue('"set description": 1' in response, response)
        self.assertTrue('"set phone": 1' in response, response)
        
        response = self.app.get(url(controller='admin', action='show'),
                                params={'serial': 'desc001'})
        self.assertTrue('phone' in response, response)
        self.assertTrue('1234560' in response, response)
        self.assertTrue('TokenDesc": "something"' in response, response)

        response = self.app.get(url(controller='admin', action='remove'),
                                params={'serial': 'desc001'})
        self.assertTrue('"value": 1' in response, response)
        
    def test_load_tokens(self):
        '''
        '''



# -*- coding: utf-8 -*-
#
#  privacyIDEA
#  Jul 18, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
The SSHKeyTokenClass provides a TokenClass that stores the public
SSH key and can give the public SSH key via the getotp function.
This can be used to manage SSH keys and retrieve the public ssh key
to import it to authorized keys files.
'''

import logging
from gettext import gettext as _
log = logging.getLogger(__name__)
from privacyidea.lib.util import getParam
from privacyidea.lib.log import log_with
from privacyidea.lib.tokenclass import TokenClass
import base64
import binascii


optional = True
required = False


class SSHkeyTokenClass(TokenClass):
    '''
    The SSHKeyTokenClass provides a TokenClass that stores the public
    SSH key and can give the public SSH key via the getotp function.
    This can be used to manage SSH keys and retrieve the public ssh key
    to import it to authorized keys files.
    '''

    def __init__(self, aToken):
        TokenClass.__init__(self, aToken)
        self.setType(u"sshkey")
        self.hKeyRequired = True

    @classmethod
    def getClassType(cls):
        return "sshkey"

    @classmethod
    def getClassPrefix(cls):
        return "SSHK"

    @classmethod
    @log_with(log)
    def getClassInfo(cls, key=None, ret='all'):
        '''
        getClassInfo - returns a subtree of the token definition

        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined
        :return: subsection if key exists or user defined
        :rtype: dictionary

        '''
        res = {'type': 'sshkey',
               'title': 'SSHkey Token',
               'description': _('SSH public key to be provided for use'
                                'in authorized_keys.'),
               'init': {'page': {'html': 'sshkeytoken.mako',
                                 'scope': 'enroll'},
                        'title': {'html': 'sshkeytoken.mako',
                                  'scope': 'enroll.title'},
                        },
               'config': {},
               # TODO we need to add selfservice!
               'selfservice': {},
               'policy': {},
               }
        if key is not None and key in res:
            ret = res.get(key)
        else:
            if ret == 'all':
                ret = res

        return ret

    def update(self, param):
        # check for the required parameters
        '''
        The key holds the public ssh key and this is required
        
        The key probably is of the form "ssh-rsa BASE64 comment"
        '''
        if (self.hKeyRequired is True):
            getParam(param, "otpkey", required)
            
        key_elem = param.get("otpkey").split()
        if len(key_elem) != 3:
            raise Exception("The key must consist of 'ssh-rsa BASE64 comment'")
        
        key_type = key_elem[0]
        key = key_elem[1]
        key_comment = key_elem[2]
        
        # convert key to hex
        param["otpkey"] = binascii.hexlify(base64.b64decode(key))
        self.addToTokenInfo("ssh_type", key_type)
        self.addToTokenInfo("ssh_comment", key_comment)

        # call the parents function
        TokenClass.update(self, param)
        
        
    @log_with(log)
    def get_sshkey(self):
        '''
        returns the public SSH key
        
        :return: SSH pub key
        :rtype: string
        '''

        secret = self.token.getHOtpKey()
        hex_sshkey = secret.getKey()
        sshkey = base64.b64encode(binascii.unhexlify(hex_sshkey))
        ti = self.getTokenInfo()
        key_type = ti.get("ssh_type")
        key_comment = ti.get("ssh_comment")
        return "%s %s %s" % (key_type, sshkey, key_comment)



# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  Sept 16, 2014 Cornelius Kölbel, added key generation for Token2
#
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  2014-10-03 Add getInitDetail
#             Cornelius Kölbel <cornelius@privacyidea.org>
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  LSE
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
  Description:  This file containes the mOTP token implementation:
              - http://motp.sourceforge.net/ -

  Dependencies: -

'''

from privacyidea.lib.util import getParam
from privacyidea.lib.util import required
from privacyidea.lib.util import optional
from privacyidea.lib.util import generate_otpkey

from privacyidea.lib.log import log_with

from privacyidea.lib.mOTP import mTimeOtp
from privacyidea.lib.tokenclass import TokenClass
from privacyidea.lib.apps import create_motp_url
from privacyidea.lib.reply import create_img
from pylons.i18n.translation import _

import traceback
import logging
log = logging.getLogger(__name__)


###############################################
class MotpTokenClass(TokenClass):
    '''
    implementation of the mOTP token class
    - see: http://motp.sourceforge.net/
    '''

    @classmethod
    def getClassType(cls):
        '''
        static method to return the token class identifier

        :return: fixed string
        '''

        return "motp"

    @classmethod
    def getClassPrefix(cls):
        return "LSMO"

    @classmethod
    def getClassInfo(cls, key=None, ret='all'):
        '''
        getClassInfo - returns all or a subtree of the token definition

        :param key: subsection identifier
        :type key: string

        :param ret: default return value, if nothing is found
        :type ret: user defined

        :return: subsection if key exists or user defined
        :rtype : s.o.

        '''

        res = {'type': 'motp',
               'title': 'mOTP Token',
               'description': ('mobile otp token'),
               'init': {'page': {'html': 'motptoken.mako',
                                 'scope': 'enroll', },
                        'title': {'html': 'motptoken.mako',
                                  'scope': 'enroll.title'},
                        },
               'config': {'page': {'html': 'motptoken.mako',
                                   'scope': 'config'},
                          'title': {'html': 'motptoken.mako',
                                    'scope': 'config.title', },
                          },
               'selfservice': {'enroll': {'page':
                                          {'html': 'motptoken.mako',
                                           'scope': 'selfservice.enroll'},
                                          'title':
                                          {'html': 'motptoken.mako',
                                           'scope': 'selfservice.title.enroll'
                                           },
                                          },
                               'motp_webprovision': {'page':
                                                     {'html': 'motptoken.mako',
                                                      'scope': 'selfservice.provision'},
                                                     'title':
                                                     {'html': 'motptoken.mako',
                                                      'scope': 'selfservice.title.provision'}
                                                     }
                               },
               'policy': {'selfservice': {'motp_webprovision': {'type': 'bool',
                                                                'desc': 'Enroll mOTP token via QR-Code.'}
                                          }}
               }

        if key is not None and key in res:
            ret = res.get(key)
        else:
            if ret == 'all':
                ret = res

        return ret

    @log_with(log)
    def __init__(self, a_token):
        '''
        constructor - create a token object

        :param a_token: instance of the orm db object
        :type a_token:  orm object
        '''
        TokenClass.__init__(self, a_token)
        self.setType(u"mOTP")
        self.hKeyRequired = True

        return
    
    @log_with(log)
    def getInitDetail(self, params, user=None):
        '''
        to complete the token normalisation, the response of the initialiastion
        should be build by the token specific method, the getInitDetails
        '''
        response_detail = TokenClass.getInitDetail(self, params, user)
        otpkey = self.getInfo().get('otpkey')
        if otpkey:
            tok_type = self.type.lower()
            if user is not None:
                try:
                    if tok_type.lower() in ["motp"]:
                        motp_url = create_motp_url(user.login, user.realm,
                                                   otpkey,
                                                   serial=self.getSerial())
                        response_detail["motpurl"] = {"description": _("URL for MOTP "
                                                                       "token"),
                                                      "value": motp_url,
                                                      "img": create_img(motp_url,
                                                                        width=250)
                                                      }
                   
                except Exception as ex:
                    log.error("%r" % (traceback.format_exc()))
                    log.error('failed to set motp url: %r' % ex)
                    
        return response_detail

    @log_with(log)
    def update(self, param, reset_failcount=True):
        '''
        update - process initialization parameters

        :param param: dict of initialization parameters
        :type param: dict

        :return: nothing

        '''
        if (self.hKeyRequired is True):
            genkey = int(getParam(param, "genkey", optional) or 0)
            if not param.get('keysize'):
                param['keysize'] = 16
            if 1 == genkey:
                otpKey = generate_otpkey(param['keysize'])
                del param['genkey']
            else:
                # genkey not set: check otpkey is given
                # this will raise an exception if otpkey is not present
                otpKey = getParam(param, "otpkey", required)

            param['otpkey'] = otpKey
            
        # motp token specific
        otpPin = getParam(param, "otppin", required)
        self.token.setUserPin(otpPin)

        TokenClass.update(self, param, reset_failcount)

        return

    @log_with(log)
    def checkOtp(self, anOtpVal, counter, window, options=None):
        '''
        checkOtp - validate the token otp against a given otpvalue

        :param anOtpVal: the to be verified otpvalue
        :type anOtpVal:  string

        :param counter: the counter state, that shoule be verified
        :type counter: int

        :param window: the counter +window, which should be checked
        :type window: int

        :param options: the dict, which could contain token specific info
        :type options: dict

        :return: the counter state or -1
        :rtype: int

        '''
        otplen = self.token.privacyIDEAOtpLen

        # otime contains the previous verification time
        # the new one must be newer than this!
        otime = self.token.privacyIDEACount
        secretHOtp = self.token.getHOtpKey()
        window = self.token.privacyIDEACountWindow
        secretPin = self.token.getUserPin()

        log.debug("otime %s", otime)

        mtimeOtp = mTimeOtp(secretHOtp, secretPin, otime, otplen)
        res = mtimeOtp.checkOtp(anOtpVal, window, options=options)

        if (res != -1):
            res = res - 1  # later on this will be incremented by 1
        if res == -1:
            msg = "verification failed"
        else:
            msg = "verifiction was successful"

        log.debug("%s :res %r" % (msg, res))
        return res



# -*- coding: utf-8 -*-
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
several independent functions
contains utility functions
'''

from privacyidea.lib.error import ParameterError
from privacyidea.lib.config import (getFromConfig,
                                    get_privacyIDEA_config)
from privacyidea.lib.log import log_with
import binascii
from privacyidea.model.meta import Session
from privacyidea.model import Token, Realm, TokenRealm
from sqlalchemy import and_

import string
from privacyidea.lib.crypto import urandom
from privacyidea.lib.crypto import geturandom
import pkg_resources
import logging

log = logging.getLogger(__name__)
import re
ENCODING = "utf-8"

SESSION_KEY_LENGTH = 32

optional = True
required = False


def get_version_number():
    '''
    returns the privacyidea version
    '''
    version = "unknown"
    try:
        version = pkg_resources.get_distribution("privacyidea").version
    except:
        pass
    return version


def get_version():
    '''
    This returns the version, that is displayed in the WebUI and
    self service portal.
    '''
    version = get_version_number()
    return "privacyIDEA %s" % version


def get_copyright_info():
    '''
    This returns the copyright information displayed in the WebUI and
    selfservice portal.
    '''
    return "privacyidea.org"


def getParam(param, which, optional=True):
    """
    getParam()
    input:
     - param (hash set): the set, which contains all parameters
     - which (lteral): the entry lookup
     - optional (boolean): defines if this parameter is optional or not
                 - an exception is thrown if the parameter is required
                 - otherwise: nothing done!

    return:
     - the value (literal) of the parameter if exists or nothing
       in case the parameter is optional, otherwise throw an exception
    """
    ret = None

    if which in param:
        ret = param[which]
    else:
        if (optional is False):
            raise ParameterError("Missing parameter: %r" % which, id=905)

    return ret


@log_with(log)
def getLowerParams(param):
    ret = {}
    for key in param:
        lkey = key.lower()
        # strip the session parameter!
        if "session" != lkey:
            lval = param[key]
            ret[lkey] = lval
    return ret


def getUserRealm(param, optional):
    user = ""
    realm = ""
    return (user, realm)


def uniquify(doubleList):
    # uniquify the realm list
    uniqueList = []
    for e in doubleList:
        if e.lower() not in uniqueList:
            uniqueList.append(e.lower())

    return uniqueList


def generate_otpkey(key_size=20):
    '''
    generates the HMAC key of keysize. Should be 20 or 32
    The key is returned as a hexlified string
    '''
    log.debug("generating key of size %s" % key_size)
    return binascii.hexlify(geturandom(key_size))


def generate_password(size=6, characters=string.ascii_lowercase +
                      string.ascii_uppercase + string.digits):
    return ''.join(urandom.choice(characters) for _x in range(size))


def check_session(request):
    '''
    This function checks the session cookie
    
    '''
    res = True
    cookie = None
    session = None
    log.debug(request.path.lower())
    # GUI function get not passed the session key.
    if request.path.lower() in ["/manage",
                                "/manage/",
                                "/manage/custom-style.css",
                                "/selfservice/custom-style.css",
                                "/",
                                "/selfservice/",
                                "/selfservice/index",
                                "/manage/tokenview",
                                "/manage/userview",
                                "/manage/policies",
                                "/manage/audittrail",
                                "/manage/machines",
                                "/selfservice/load_form",
                                "/selfservice/assign",
                                "/selfservice/resync",
                                "/selfservice/reset",
                                "/selfservice/getotp",
                                "/selfservice/disable",
                                "/selfservice/enable",
                                "/selfservice/unassign",
                                "/selfservice/delete",
                                "/selfservice/setpin",
                                "/selfservice/setmpin",
                                "/selfservice/history",
                                "/selfservice/webprovisionoathtoken",
                                "/selfservice/activateqrtoken",
                                "/selfservice/webprovisiongoogletoken"
                                ]:
        log.info('nothing to check')
    else:
        if request.cookies.get('privacyidea_session') is not None:
            try:
                cookie = request.cookies.get('privacyidea_session')[0:40]
                session = request.params.get('session')[0:40]
            except Exception as e:
                log.warning("failed to check selfservice session: %r" % e)
                res = False
            log.info("session: %s" % session)
            log.info("cookie:  %s" % cookie)
            if session is None or session != cookie:
                log.error("The request %s did not pass a valid session!" %
                          request.url)
                res = False
    return res


def remove_session_from_param(param):
    '''
    Some low level functions like the userlisting do not like to have a
    session parameter in the param dictionary.
    So we remove the session from the params.
    '''
    return_param = {}
    for key in param.keys():
        if "session" != key.lower():
            return_param[key] = param[key]

    return return_param


# -------------------------------------------------------------------------
# Client overwriting stuff
#

def get_client_from_request(request):
    '''
    This function returns the client as it is passed in the HTTP Request.
    This is the very HTTP client, that contacts the privacyIDEA server.
    '''
    client = "unknown"
    try:
        client = request.environ.get('REMOTE_ADDR', None)
        log.debug("got the client %s" % client)
    except Exception as exx:
        log.warning("%r" % exx)
    return client


@log_with(log)
def get_client_from_param(param):
    '''
    This function returns the client, that is passed with
    the GET/POST parameter "client"
    '''
    client = getParam(param, "client", optional)
    return client


def get_client():
    '''
    This function returns the client.

    It first tries to get the client as it is passed as the HTTP Client
    via REMOTE_ADDR.

    If this client Address is in a list, that is allowed to overwrite its
    client address (like e.g. a FreeRADIUS server, which will always pass
    the FreeRADIUS address but not the address of the
    RADIUS client) it checks for the existence of the client parameter.
    '''
    # FIXME pylons dependency!
    from pylons import request
    may_overwrite = []
    over_client = getFromConfig("mayOverwriteClient", "")
    log.debug("config entry mayOverwriteClient: %s" % over_client)
    try:
        may_overwrite = [c.strip() for c in over_client.split(',')]
    except Exception as e:
        log.warning("evaluating config entry 'mayOverwriteClient': %r" % e)

    client = get_client_from_request(request)
    log.debug("got the original client %s" % client)

    if client in may_overwrite or client is None:
        log.debug("client %s may overwrite!" % client)
        if get_client_from_param(request.params):
            client = get_client_from_param(request.params)
            log.debug("client overwritten to %s" % client)

    log.debug("returning %s" % client)
    return client


def normalize_activation_code(activationcode,
                              upper=True,
                              convert_o=True,
                              convert_0=True):
    '''
    This normalizes the activation code.
    1. lower letters are capitaliezed
    2. Oh's in the last two characters are turned to zeros
    3. zeros in in the first-2 characters are turned to Ohs
    '''
    if upper:
        activationcode = activationcode.upper()
    if convert_o:
        activationcode = activationcode[:-2] + activationcode[-2:].replace("O",
                                                                           "0")
    if convert_0:
        activationcode = activationcode[:-2].replace("0",
                                                     "O") + activationcode[-2:]

    return activationcode


def is_valid_fqdn(hostname, split_port=False):
    '''
    Checks if the hostname is a valid FQDN
    '''
    if split_port:
        hostname = hostname.split(':')[0]
    if len(hostname) > 255:
        return False
    if hostname[-1:] == ".":
        # strip exactly one dot from the right, if present
        hostname = hostname[:-1]
    allowed = re.compile("(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in hostname.split("."))


def remove_empty_lines(doc):
    '''
        remove empty lines from the input document

        @param doc: documemt containing long multiline text
        @type  doc: string

        @return: data without empty lines
        @rtype:  string

    '''
    data = '\n'.join([line for line in doc.split('\n') if line.strip() != ''])
    return data

#
# Modhex calculations for Yubikey
#
hexHexChars = '0123456789abcdef'
modHexChars = 'cbdefghijklnrtuv'

hex2ModDict = dict(zip(hexHexChars, modHexChars))
mod2HexDict = dict(zip(modHexChars, hexHexChars))


def modhex_encode(s):
    return ''.join(
        [hex2ModDict[c] for c in s.encode('hex')]
    )
# end def modhex_encode


def modhex_decode(m):
    return ''.join(
        [mod2HexDict[c] for c in m]
    ).decode('hex')
# end def modhex_decode


def checksum(msg):
    crc = 0xffff
    for i in range(0, len(msg) / 2):
        b = int(msg[i * 2] + msg[(i * 2) + 1], 16)
        crc = crc ^ (b & 0xff)
        for _j in range(0, 8):
            n = crc & 1
            crc = crc >> 1
            if n != 0:
                crc = crc ^ 0x8408
    return crc


@log_with(log)
def getRealmBox():
    '''
    returns the config value of selfservice.realmbox.
    if True, the realmbox in the selfservice login will be displayed.
    if False, the realmbox will not be displayed and the user needs to login
    via user@realm
    '''
    rb_string = "privacyidea.selfservice.realmbox"
    conf = get_privacyIDEA_config()
    if rb_string in conf:
        log.debug("read setting: %r" % conf.get(rb_string))
        return "True" == conf.get(rb_string)
    else:
        return False


@log_with(log)
def get_login_help():
    '''
    returns the config value of "login.help".
    if True, the help button is displayed on the login screen.
    '''
    lh_string = "privacyidea.login.help"
    conf = get_privacyIDEA_config()
    if lh_string in conf:
        log.debug("read setting: %r" % conf.get(lh_string))
        return "True" == conf.get(lh_string)
    else:
        return True


# ######################### base token ##############################


def get_token_in_realm(realm, active=True):
    '''
    This returns the number of tokens in one realm.

    You can either query only active token or also disabled tokens.
    '''
    if active:
        sqlQuery = Session.query(TokenRealm,
                                 Realm,
                                 Token).filter(and_(TokenRealm.realm_id == Realm.id,
                                                    Realm.name == u'' + realm,
                                                    Token.privacyIDEAIsactive == True,
                                                    TokenRealm.token_id == Token.privacyIDEATokenId)).count()
    else:
        sqlQuery = Session.query(TokenRealm,
                                 Realm).filter(and_(TokenRealm.realm_id == Realm.id,
                                                    Realm.name == realm)).count()
    return sqlQuery

####################################################################



# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  LSE
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""The application's Globals object"""
import threading
import copy
import logging

from privacyidea.lib.security.provider import SecurityProvider


log = logging.getLogger(__name__)

class Globals(object):

    """Globals acts as a container for objects available throughout the
    life of the application

    """

    def __init__(self):
        """One instance of Globals is created during application
        initialization and is available during requests via the
        'app_globals' variable

        """
        self.rwl = RWLock()
        self.rwl2 = RWLock()
        self.resolverLock = RWLock()
        self.rcount = 0

        self.config = {}
        self.config_incomplete = False
        self.configLock = RWLock()
        secLock = RWLock()

        self.tokenprefixes = {}
        self.tokenclasses = {}
        self.security_provider = SecurityProvider(secLock)

        self.resolver_clazzes = {}
        self.resolver_types = {}

    def setResolverClasses(self, resolver_clazzes=None):
        '''
        setter to hold the reference to all resolver class objects
        '''
        if resolver_clazzes is not None:
            self.resolver_clazzes = resolver_clazzes

    def getResolverClasses(self):
        return self.resolver_clazzes

    def setResolverTypes(self, resolver_types=None):
        """
        setter to hold the reference to all resolver class names
        """
        if resolver_types is not None:
            self.resolver_types = resolver_types

    def getResolverTypes(self):
        return self.resolver_types


    def getConfig(self):
        '''
            retrieve (the deep copy of) the actual config
        '''
        self.configLock.acquire_read()
        try:
            config = copy.deepcopy(self.config)
        finally:
            self.configLock.release()
        return config

    def setTokenclasses(self, tcl):
        self.tokenclasses = tcl
        return
    
    def set_applications(self, apps):
        self.applications = apps
        return
    
    def getTokenclasses(self):
        return self.tokenclasses

    def setTokenprefixes(self, tpl):
        self.tokenprefixes = tpl
        return

    def getTokenprefixes(self):
        return self.tokenprefixes


    def setConfig(self, config, replace=False):
        '''
            set the app global config for privacyidea
        '''
        err = None
        self.configLock.acquire_write()
        try:
            ty = type(config).__name__
            if ty != 'dict':
                self.configLock.release()
                err = 'cannot set global config from object ' + ty

            else:
                conf = copy.deepcopy(config)
                if replace == True:
                    self.config = conf
                else:
                    self.config.update(conf)
        finally:
            self.configLock.release()
            if err is not None:
                raise Exception(err)
        return

    def isConfigComplet(self):
        ret = True
        self.configLock.acquire_read()
        try:
            ret = self.config_incomplete
        finally:
            self.configLock.release()
        return  ret

    def setConfigIncomplete(self, val=False):
        '''
            set the app global config for privacyidea
        '''
        self.configLock.acquire_write()
        try:
            self.config_incomplete = val
        finally:
            self.configLock.release()
        return


    def delConfig(self, conf):
        '''
            delete one entry in the appl_globals
        '''
        self.configLock.acquire_write()
        try:
            ty = type(conf).__name__

            if ty == 'list' or ty == 'dict':
                for k in conf:
                    if self.config.has_key(k):
                        del self.config[k]
            elif ty == 'str' or ty == 'unicode':
                if self.config.has_key(conf):
                    del self.config[conf]
        finally:
            self.configLock.release()
        return


    def getLock(self):
        return self.rwl

    def setConfigReadLock(self):
        self.rcount = self.rcount + 1
        self.rwl2.acquire_read()
        return self.rcount

    def setConfigWriteLock(self):
        self.rcount = self.rcount + 1
        self.rwl2.acquire_write()
        return self.rcount

    def releaseConfigLock(self):
        self.rcount = self.rcount - 1
        self.rwl2.release()
        return self.rcount


###
#Python offers a number of useful synchronization primitives in the threading and Queue modules.
#One that is missing, however, is a simple reader-writer lock (RWLock). A RWLock allows improved
#concurrency over a simple mutex, and is useful for objects that have high read-to-write ratios
#like database caches.
#
#Surprisingly, I haven t been able to find any implementation of these semantics, so I rolled my
#own in a module rwlock.py to implement a RWLock class, along with lock promotion/demotion. Hopefully
#it can be added to the standard library threading module.
#This code is hereby placed in the public domain.
#
#    from
#    http://majid.info/blog/a-reader-writer-lock-for-python/
#
#
#   Simple reader-writer locks in Python
#    Many readers can hold the lock XOR one and only one writer
#
#
###
#
#version = """$Id: 04-1.html,v 1.3 2006/12/05 17:45:12 majid Exp $"""

class RWLock:
    """
    A simple reader-writer lock Several readers can hold the lock
    simultaneously, XOR one writer. Write locks have priority over reads to
    prevent write starvation.
    """
    def __init__(self):
        self.rwlock = 0
        self.writers_waiting = 0
        self.monitor = threading.Lock()
        self.readers_ok = threading.Condition(self.monitor)
        self.writers_ok = threading.Condition(self.monitor)
    def acquire_read(self):
        """Acquire a read lock. Several threads can hold this typeof lock.
        It is exclusive with write locks.
        """
        self.monitor.acquire()
        while self.rwlock < 0 or self.writers_waiting:
            self.readers_ok.wait()
        self.rwlock += 1
        self.monitor.release()
    def acquire_write(self):
        """Acquire a write lock. Only one thread can hold this lock, and
            only when no read locks are also held.
        """
        self.monitor.acquire()
        while self.rwlock != 0:
            self.writers_waiting += 1
            self.writers_ok.wait()
            self.writers_waiting -= 1
        self.rwlock = -1
        self.monitor.release()
    def promote(self):
        """Promote an already-acquired read lock to a write lock
        WARNING: it is very easy to deadlock with this method"""
        self.monitor.acquire()
        self.rwlock -= 1
        while self.rwlock != 0:
            self.writers_waiting += 1
            self.writers_ok.wait()
            self.writers_waiting -= 1
        self.rwlock = -1
        self.monitor.release()
    def demote(self):
        """Demote an already-acquired write lock to a read lock"""
        self.monitor.acquire()
        self.rwlock = 1
        self.readers_ok.notifyAll()
        self.monitor.release()
    def release(self):
        """Release a lock, whether read or write."""
        self.monitor.acquire()
        if self.rwlock < 0:
            self.rwlock = 0
        else:
            self.rwlock -= 1
        wake_writers = self.writers_waiting and self.rwlock == 0
        wake_readers = self.writers_waiting == 0
        self.monitor.release()
        if wake_writers:
            self.writers_ok.acquire()
            self.writers_ok.notify()
            self.writers_ok.release()
        elif wake_readers:
            self.readers_ok.acquire()
            self.readers_ok.notifyAll()
            self.readers_ok.release()



# -*- coding: utf-8 -*-
#
#  privacyIDEA
#  (c) 2014 Cornelius Kölbel, cornelius@privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
This file is part of the privacyidea service
'''
import logging

from pylons import request, response, config, tmpl_context as c

from privacyidea.lib.base import BaseController
from privacyidea.lib.token import get_token_type_list, getTokenType

from privacyidea.lib.util import getParam
from privacyidea.weblib.util import get_client
from privacyidea.lib.user import getUserFromRequest
from privacyidea.lib.reply import sendResult, sendError

from privacyidea.model.meta import Session
from privacyidea.lib.policy import PolicyClass
from privacyidea.lib.policy import PolicyException
from privacyidea.lib.error import ParameterError
from privacyidea.lib.config import get_privacyIDEA_config

from privacyidea.lib.machine import delete as delete_machine
from privacyidea.lib.machine import create as create_machine
from privacyidea.lib.machine import show as show_machine
from privacyidea.lib.machine import addtoken
from privacyidea.lib.machine import deltoken
from privacyidea.lib.machine import showtoken
from privacyidea.lib.machine import get_token_apps
from privacyidea.lib.machine import addoption
from privacyidea.lib.machine import deloption

import traceback
import webob
import json

from privacyidea.lib.log import log_with

log = logging.getLogger(__name__)

optional = True
required = False


class MachineController(BaseController):

    @log_with(log)
    def __before__(self, action, **params):
        '''
        '''
        try:
            c.audit['success'] = False
            c.audit['client'] = get_client()
            self.Policy = PolicyClass(request, config, c,
                                      get_privacyIDEA_config(),
                                      tokenrealms=request.params.get('serial'),
                                      token_type_list=get_token_type_list())
            self.set_language()

            self.before_identity_check(action)

            Session.commit()
            return request

        except webob.exc.HTTPUnauthorized as acc:
            # the exception, when an abort() is called if forwarded
            log.info("%r: webob.exception %r" % (action, acc))
            log.info(traceback.format_exc())
            Session.rollback()
            Session.close()
            raise acc

        except Exception as exx:  # pragma: no cover
            log.error("exception %r" % (action, exx))
            log.error(traceback.format_exc())
            Session.rollback()
            Session.close()
            return sendError(response, exx, context='before')

        finally:
            pass

    @log_with(log)
    def __after__(self, action, **params):
        '''
        '''
        params = {}

        try:
            params.update(request.params)
            c.audit['administrator'] = getUserFromRequest(request).get("login")
            if 'serial' in params:
                    c.audit['serial'] = request.params['serial']
                    c.audit['token_type'] = getTokenType(params.get('serial'))

            self.audit.log(c.audit)

            Session.commit()
            return request

        except Exception as e:  # pragma: no cover
            log.error("unable to create a session cookie: %r" % e)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, e, context='after')

        finally:
            Session.close()

    @log_with(log)
    def create(self, action, **params):
        '''
        Create a new client machine entry

        :param name: the unique name of the machine (required).
                     Can be the FQDN.
        :param desc: description of the machine
        :param ip: The IP address of the machine (required)
        :param decommission: A date when the machine will not be valid anymore

        :return: True or False if the creation was successful.
        '''
        try:
            res = False
            param = {}
            # check machine authorization
            self.Policy.checkPolicyPre('machine', 'create')

            param.update(request.params)
            machine_name = getParam(param, "name", required)
            ip = getParam(param, "ip", required)
            desc = getParam(param, "desc", optional)
            decommission = getParam(param, "decommission", optional)
            machine = create_machine(machine_name,
                                     ip=ip,
                                     desc=desc,
                                     decommission=decommission)
            if machine:
                res = True
            Session.commit()
            c.audit["success"] = True
            return sendResult(response, res, 1)

        except PolicyException as pe:
            log.error("policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe))

        except Exception as exx:  # pragma: no cover
            log.error("failed: %r" % exx)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(exx), 0)

        finally:
            Session.close()
            
    @log_with(log)
    def delete(self, action, **params):
        '''
        Delete an existing client machine entry
        
        :param name: the unique name of the machine
        
        :return: value is either true (success) or false (fail)
        '''
        try:
            res = {}
            param = {}
            # check machine authorization
            self.Policy.checkPolicyPre('machine', 'delete')
            param.update(request.params)
            machine_name = getParam(param, "name", required)
            res = delete_machine(machine_name)
            Session.commit()
            c.audit["success"] = True
            return sendResult(response, res, 1)

        except PolicyException as pe:
            log.error("policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe))

        except Exception as exx:  # pragma: no cover
            log.error("failed: %r" % exx)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(exx), 0)

        finally:
            Session.close()

    @log_with(log)
    def show(self, action, **params):
        '''
        Returns a list of the client machines.
        
        :param name: Optional parameter to only show this single machine
        
        :return: JSON details
        '''
        try:
            res = {}
            param = {}
            # check machine authorization
            self.Policy.checkPolicyPre('machine', 'show')
            param.update(request.params)
            machine_name = getParam(param, "name", optional)
            
            res = show_machine(machine_name)
            Session.commit()
            c.audit["success"] = True
            return sendResult(response, res, 1)

        except PolicyException as pe:
            log.error("policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe))

        except Exception as exx:  # pragma: no cover
            log.error("failed: %r" % exx)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(exx), 0)

        finally:
            Session.close()

    @log_with(log)
    def addtoken(self, action, **params):
        '''
        Add a token and a application to a machine
        
        :param name: Name of the machine
        :param serial: serial number of the token
        :param application: name of the application
        :param option_*: parameter is passed as additional option to
                         the machinetoken to be stored in
                         machine_t_options table.
                         In case of LUKS application this can be
                         "option_slot"
        '''
        try:
            res = False
            param = {}
            options = {}
            # check machine authorization
            self.Policy.checkPolicyPre('machine', 'addtoken')
            param.update(request.params)
            machine_name = getParam(param, "name", required)
            serial = getParam(param, "serial", required)
            application = getParam(param, "application", required)
            
            if application.lower() not in config.get("applications").keys():
                log.error("Unknown application %r. Available applications: "
                          "%r" % (application,
                                  config.get("applications").keys()))
                raise Exception("Unkown application!")
            
            for p in param.keys():
                if p.startswith("option_"):
                    options[p] = param.get(p)
            
            mt = addtoken(machine_name,
                          serial,
                          application,
                          options=options)
            if mt:
                res = True
            Session.commit()
            c.audit["success"] = True
            return sendResult(response, res, 1)

        except PolicyException as pe:
            log.error("policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe))

        except Exception as exx:  # pragma: no cover
            log.error("failed: %r" % exx)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(exx), 0)

        finally:
            Session.close()

    @log_with(log)
    def addoption(self, action, **params):
        '''
        Add an option to a machinetoken definition
        
        :param mtid: id of the machine token definition
        :param name: machine_name
        :param serial: serial number of the token
        :param application: application
        :param option_*: name of the option.
                         In case of LUKS application this can be
                         "option_slot"
                         
        You either need to provide the machine-token-id (mtid) directly or
        you need to provide the tuple (name, serial, application)
        '''
        try:
            num = -1
            param = {}
            options = {}
            # check machine authorization
            self.Policy.checkPolicyPre('machine', 'addtoken')
            param.update(request.params)
            mtid = getParam(param, "mtid", optional)
            machine_name = getParam(param, "name", optional)
            serial = getParam(param, "serial", optional)
            application = getParam(param, "application", optional)
            if not (mtid or (machine_name and serial and application)):
                raise ParameterError("You need to specify either mtid or"
                                     "the tuple name, serial, application",
                                     id=201)
               
            for p in param.keys():
                if p.startswith("option_"):
                    options[p] = param.get(p)
            
            num = addoption(mtid,
                            name=machine_name,
                            serial=serial,
                            application=application,
                            options=options)
            Session.commit()
            c.audit["success"] = True
            return sendResult(response, num, 1)

        except PolicyException as pe:
            log.error("policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe))

        except Exception as exx:  # pragma: no cover
            log.error("failed: %r" % exx)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(exx), 0)

        finally:
            Session.close()

    @log_with(log)
    def deloption(self, action, **params):
        '''
        Delete an option from a machinetoken definition
        
        :param mtid: id of the machine token definition
        :param name: machine_name
        :param serial: serial number of the token
        :param application: application
        :param key: key of the option to delete
                         
        You either need to provide the machine-token-id (mtid) directly or
        you need to provide the tuple (name, serial, application)
        '''
        try:
            num = -1
            param = {}
            options = {}
            # check machine authorization
            self.Policy.checkPolicyPre('machine', 'deltoken')
            param.update(request.params)
            mtid = getParam(param, "mtid", optional)
            machine_name = getParam(param, "name", optional)
            serial = getParam(param, "serial", optional)
            option_key = getParam(param, "key", required)
            application = getParam(param, "application", optional)
            if not (mtid or (machine_name and serial and application)):
                raise ParameterError("You need to specify either mtid or"
                                     "the tuple name, serial, application",
                                     id=201)
            num = deloption(mtid,
                            name=machine_name,
                            serial=serial,
                            application=application,
                            key=option_key)
            Session.commit()
            c.audit["success"] = True
            return sendResult(response, num, 1)

        except PolicyException as pe:
            log.error("policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe))

        except Exception as exx:  # pragma: no cover
            log.error("failed: %r" % exx)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(exx), 0)

        finally:
            Session.close()

    @log_with(log)
    def deltoken(self, action, **params):
        '''
        delete a token and a application from a machine
        
        :param name: Name of the machine
        :param serial: serial number of the token
        :param application: name of the application
        '''
        try:
            res = False
            param = {}
            # check machine authorization
            self.Policy.checkPolicyPre('machine', 'deltoken')
            param.update(request.params)
            machine_name = getParam(param, "name", required)
            serial = getParam(param, "serial", required)
            application = getParam(param, "application", required)
            
            res = deltoken(machine_name, serial, application)
            Session.commit()
            c.audit["success"] = True
            return sendResult(response, res, 1)

        except PolicyException as pe:
            log.error("policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe))

        except Exception as exx:  # pragma: no cover
            log.error("failed: %r" % exx)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(exx), 0)
        finally:
            Session.close()

    @log_with(log)
    def showtoken(self, action, **params):
        '''
        show a token and a application from a machine
        This function is used in the flexigrid management
        
        :param name: Name of the machine
        :param serial: serial number of the token
        :param application: name of the application
        :param flexi: if set to 1, we do return flexigrid input
        '''
        try:
            res = False
            param = {}
            # check machine authorization
            self.Policy.checkPolicyPre('machine', 'showtoken')
            param.update(request.params)
            machine_name = getParam(param, "name", optional)
            serial = getParam(param, "serial", optional)
            application = getParam(param, "application", optional)
            # if set, this should be returned for flexigrid
            flexi = getParam(param, "flexi", optional)

            res = showtoken(machine_name,
                            serial,
                            application,
                            flexi=flexi,
                            params=param)
            Session.commit()
            c.audit["success"] = True
            if flexi:
                response.content_type = 'application/json'
                return json.dumps(res, indent=3)
            else:
                return sendResult(response, res, 1)

        except PolicyException as pe:
            log.error("policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe))

        except Exception as exx:  # pragma: no cover
            log.error("failed: %r" % exx)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(exx), 0)

        finally:
            Session.close()

    @log_with(log)
    def gettokenapps(self, action, **params):
        '''
        returns the apps and the authentication information
        for the given machine.

        If an application is given only the authentication item for
        this application is returned. otherwise the application items
        for all applications are returned.

        TODO: Authenticate the client machine

        :param name: the machine name -
                    otherwise the machine is identified by the IP
        :type name: string, optional
        :param application: the name of the application. If the application is
                            given the response will also conatin the auth_item
        :type application: string, optional
        :param serial: The serial number of the token
        
        :return: dictionary with machine and authentication information
        
        
        {"total": 2,
         "machines":  { 1 : {"application": "luks",
                             "auth_item": {"challenge": "hex...",
                                           "response":: "hex..."},
                             "id": 1,
                             "ip": 10.0.0.1,
                             "machine_id": 1,
                             "machinename": "computerABC",
                             "serial": "<serialnumber of token>",
                             "token_id": 1
                            },
                        2 : ...
                      }
        }
        '''
        try:
            res = False
            param = {}
            self.Policy.checkPolicyPre('machine', 'gettokenapps')
            param.update(request.params)
            machine_name = getParam(param, "name", optional)
            application = getParam(param, "application", optional)
            serial = getParam(param, "serial", optional)
            # In case of LUKS and Yubikey we need the client to
            # pass the challenge as the challenge is the Password
            challenge = getParam(param, "challenge", optional)
            client_ip = get_client()
            
            if application:
                if application not in config.get("applications").keys():
                    log.error("Unknown application %r. Available applications:"
                              "%r" % (application,
                                      config.get("applications").keys()))
            application_module = config.get("applications").get(application)
            
            res = get_token_apps(machine=machine_name,
                                 application=application,
                                 application_module=application_module,
                                 serial=serial,
                                 client_ip=client_ip,
                                 challenge=challenge)
            Session.commit()
            c.audit["success"] = True
            return sendResult(response, res, 1)

        except PolicyException as pe:
            log.error("policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe))

        except Exception as exx:  # pragma: no cover
            log.error("failed: %r" % exx)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(exx), 0)

        finally:
            Session.close()
            
    @log_with(log)
    def getapplications(self, action, **params):
        '''
        Returns a list of available applications
        '''
        try:
            return sendResult(response, config.get("applications").keys())
        
        except PolicyException as pe:
            log.error("policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe))

        except Exception as exx:  # pragma: no cover
            log.error("failed: %r" % exx)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(exx), 0)

        finally:
            Session.close()
            


# -*- coding: utf-8 -*-
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''     
  Description:  functional tests
                
  Dependencies: -

'''

import logging
from privacyidea.tests import TestController, url

log = logging.getLogger(__name__)

class TestRemoteToken(TestController):

    def setUp(self):
        '''
        Overwrite the deleting of the realms!

        If the realms are deleted also the table TokenRealm gets deleted
        and we loose the information how many tokens are within a realm!
        '''
        return

    def tearDown(self):
        ''' Overwrite parent tear down, which removes all realms '''
        return

    ### define Admins

    def test_00_000(self):
        '''
        Init the tests....
        '''
        self.deleteAllPolicies()
        self.deleteAllTokens()

        self.__createResolvers__()
        self.__createRealms__()

        return

    def test_00_create_remote_token(self):
        # local token
        param_local_1 = {
                         "serial"  : "LSPW1",
                      "type"    : "pw",
                      "otpkey"  : "123456",
                      "otppin"  : "",
                      "user"    : "",
                      "pin"     : "pin",
                         }
        param_local_2 = {
                         "serial"  : "LSPW2",
                      "type"    : "pw",
                      "otpkey"  : "234567",
                      "otppin"  : "",
                      "user"    : "",
                      "pin"     : "pin",
                         }

        # The token with the remote PIN
        parameters1 = {
                      "serial"  : "LSRE001",
                      "type"    : "remote",
                      "otpkey"  : "1234567890123456",
                      "otppin"  : "",
                      "user"    : "remoteuser",
                      "pin"     : "pin",
                      "description" : "RemoteToken1",
                      'remote.server' : 'http://127.0.0.1:5001',
                      'remote.local_checkpin' : 0,
                      'remote.serial' : 'LSPW1',
                      }

        # the token with the local PIN
        parameters2 = {
                      "serial"  : "LSRE002",
                      "type"    : "remote",
                      "otpkey"  : "1234567890123456",
                      "otppin"  : "",
                      "user"    : "localuser",
                      "pin"     : "pin",
                      "description" : "RemoteToken2",
                      'remote.server' : 'http://127.0.0.1:5001',
                      'remote.local_checkpin' : 1,
                      'remote.serial' : 'LSPW2',
                      }


        response = self.app.get(url(controller='admin', action='init'),
                                params=param_local_1)
        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='init'),
                                params=param_local_2)
        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters1)
        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters2)
        self.assertTrue('"value": true' in response, response)

        response = self.app.get(url(controller='admin', action='set'),
                                params={'serial':'LSPW1', 'pin':'lspw1'})
        self.assertTrue('"set pin": 1' in response, response)

        response = self.app.get(url(controller='admin', action='set'),
                                params={'serial':'LSPW2', 'pin':''})
        self.assertTrue('"set pin": 1' in response, response)

        response = self.app.get(url(controller='admin', action='set'),
                                params={'serial':'LSRE001', 'pin':'local'})
        self.assertTrue('"set pin": 1' in response, response)

        response = self.app.get(url(controller='admin', action='set'),
                                params={'serial':'LSRE002', 'pin':'local'})
        self.assertTrue('"set pin": 1' in response, response)

        return

    def test_02_check_token_local_pin(self):
        '''
        Checking if token with local PIN works

        To successfully test the remote token, the paster must run locally.
        '''

        parameters = { "serial" : "LSPW2",
                       "pass" : "234567" }
        response = self.app.get(url(controller='validate', action='check_s'),
                                params=parameters)
        log.error(response)
        assert '"value": true' in response

        parameters = {"user": "localuser", "pass": "local234567"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)
        log.error(response)

        assert '"value": true' in response

        return


    def test_03_check_token_remote_pin(self):
        '''
        Checking if remote PIN works
        '''
        parameters = {"user": "remoteuser", "pass": "lspw1123456"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        self.assertTrue('"value": true' in response, response)

        return

    def test_04_check_token_local_pin_fail(self):
        '''
        Checking if a missing local PIN will fail
        '''

        parameters = {"user": "localuser", "pass": "234567"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        assert '"value": false' in response

        return

    def test_05_check_token_local_pin_fail2(self):
        '''
        Checking if a wrong local PIN will fail
        '''

        parameters = {"user": "localuser", "pass": "lspw1234567"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        assert '"value": false' in response

    def test_06_check_token_remote_pin_fail(self):
        '''
        Checking if a missing remote PIN will fail
        '''

        parameters = {"user": "remoteuser", "pass": "123456"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        assert '"value": false' in response

        return

    def test_06_check_token_remote_pin_fail2(self):
        '''
        Checking if a wrong remote PIN will fail
        '''

        parameters = {"user": "remoteuser", "pass": "local123456"}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        assert '"value": false' in response

        return

    def create_local_tokens(self, serial):

        serial = "LSP%s" % serial

        # local token
        param_local_1 = {
                       "serial"  : serial,
                      "type"    : "spass",
                      "otpkey"  : "123456",
                      "otppin"  : "",
                      "user"    : "",
                      "pin"     : "pin",
                         }

        response = self.app.get(url(controller='admin', action='init'),
                                params=param_local_1)
        self.assertTrue('"value": true' in response, response)
        return serial

    def delete_tokens(self, serial):
        ##
        param_local_1 = {'serial' : serial}

        response = self.app.get(url(controller='admin', action='remove'),
                                params=param_local_1)
        self.assertTrue('"value": 1' in response, response)
        return

    def test_07_fix_12061(self):
        '''
        ticket 12061: timeout with remote tokens: many tokens + unicode pins
        '''

        sqlconnect = self.appconf.get('sqlalchemy.url')
        log.debug('current test against %s' % (sqlconnect))

        from privacyidea.model import token_table
        for column in token_table.columns:
            log.debug("Column Table name: %s : %s : %r"
                      % (column.name, column.type, column.index))
            if column.name == 'privacyIDEATokenSerialnumber':
                assert column.index == True

        serials = []

        for i in range(1, 90):
            serial = self.create_local_tokens('tok_%d' % i)
            serials.append(serial)

        serial = serials[0]
        rserial = "%s_remote" % serial
        serials.append(rserial)

        parameters1 = {
                      "serial"  : rserial,
                      "type"    : "remote",
                      "otpkey"  : "1234567890123456",
                      "otppin"  : "",
                      "user"    : "root",
                      "pin"     : "",
                      "description" : "RemoteToken",
                      'remote.server' : 'http://127.0.0.1:5001',
                      'remote.local_checkpin' : 0,
                      'remote.serial' : serial,
                      }

        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters1)
        self.assertTrue('"value": true' in response, response)

        for offset in range(1, 20):
            pin = "pin_"
            for i in range(1, 100):
                pin = "%s%s" % (pin, unichr(0x28 * offset + i))

            params = { 'serial' : serial, 'pin' : pin}
            response = self.app.get(url(controller='admin', action='set'),
                                    params=params)
            self.assertTrue('"set pin": 1' in response, response)

            params = {'user':'root', 'pass' : pin}
            response = self.app.get(url(controller='validate',
                                        action='check'),
                                    params=params)
            self.assertTrue('"value": true' in response, response)

        for serial in serials:
            self.delete_tokens(serial)

        return

#eof###########################################################################




# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
'''         
  Description: Helper functions
                Consists of functions to typically be used within templates, but also
                available to Controllers. This module is available to templates as 'h'. 
      
  Dependencies: -

'''

# Import helpers as desired, or define your own, ie:
#from webhelpers.html.tags import checkbox, password
import logging


#CKO: depending on the version there are two different possible locations...
# see http://stackoverflow.com/questions/2219316/pylons-webhelpers-missing-secure-form-module
from privacyidea.lib.error import ParameterError


log = logging.getLogger(__name__)

def getParam(param, which, optional):
    if param.has_key(which):
        return  param[which]
    else:
        if (optional == False):
            raise ParameterError("Missing parameter: %r" % which, id=905)
    return
